"""LightsOut MaxSAT-based decoder for the hexagonal color code."""
from __future__ import annotations

import datetime
import json
import subprocess
from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, Any

import numpy as np
from z3 import Bool, Not, Optimize, Xor, simplify

from . import code_from_string

if TYPE_CHECKING:  # pragma: no cover
    import numpy.typing as npt
    from z3 import ModelRef

    from . import ColorCode


@dataclass
class LightsOut:
    """Lights out problem representation."""

    lights_to_switches: dict[int, list[int]]
    switches_to_lights: dict[int, list[int]]
    switch_vars: list[Bool] | None = None
    helper_vars: dict[int, list[Bool]] = field(default_factory=dict)
    optimizer: Optimize = field(default_factory=Optimize)

    def preconstruct_parity_constraint(self, light: int, indices: list[int]) -> None:
        """Preconstruct the parity constraints for a light.

        Adds all constraint to the optimizer that are independent of the value of the light.
        """
        helper_vars = self.helper_vars[light]

        assert self.switch_vars is not None
        for i in range(1, len(indices) - 1):
            constraint = Xor(self.switch_vars[indices[i]], helper_vars[i]) == helper_vars[i - 1]
            self.optimizer.add(simplify(constraint))

        constraint = self.switch_vars[indices[-1]] == helper_vars[-1]
        self.optimizer.add(simplify(constraint))

    def complete_parity_constraint(self, light: int, indices: list[int], val: bool) -> None:
        """Completes the parity constraints for a light.

        Adds the constraint that is dependent on the value of the light.
        """
        helper_vars = self.helper_vars[light]
        assert self.switch_vars is not None
        constraint = Xor(self.switch_vars[indices[0]], helper_vars[0]) == val
        self.optimizer.add(simplify(constraint))

    def preconstruct_z3_instance(self) -> None:
        """Preconstruct the z3 instance for the lights-out problem.

        Creates all necessary variables, adds the known parts of the parity constraints.
        Soft constraints are added to the optimizer with default weights.
        """
        if self.switch_vars is None:
            self.switch_vars = [Bool(f"switch_{i}") for i in range(len(self.switches_to_lights))]

        for light, switches in self.lights_to_switches.items():
            if light not in self.helper_vars:
                self.helper_vars[light] = [Bool(f"helper_{light}_{i}") for i in range(len(switches) - 1)]
            self.preconstruct_parity_constraint(light, switches)

        for switch in self.switch_vars:
            self.optimizer.add_soft(Not(switch))

    def validate_model(self, model: ModelRef, lights: list[bool]) -> bool:
        """Validate the model by checking if pressing the switches turns off all lights."""
        assert self.switch_vars is not None
        for i, var in enumerate(self.switch_vars):
            if model[var]:
                # flip all lights that are controlled by this switch
                for light in self.switches_to_lights[i]:
                    lights[light] = not lights[light]

        return all(not light for light in lights)

    def count_switches(self, model: ModelRef) -> int:
        """Count the number of switches that are set to true."""
        assert self.switch_vars is not None
        return sum(1 for var in self.switch_vars if model[var])

    def solve(
        self, lights: list[bool], solver_path: str = "z3"
    ) -> tuple[list[int], datetime.timedelta, datetime.timedelta]:
        """Solve the lights-out problem for a given pattern.

        Assumes that the z3 instance has already been pre-constructed.
        """
        # push a new context to the optimizer
        self.optimizer.push()

        # add the problem specific constraints
        start = datetime.datetime.now()
        for light, val in enumerate(lights):
            self.complete_parity_constraint(light, self.lights_to_switches[light], val)
        constr_time = datetime.datetime.now() - start
        switches: list[int] = []
        if solver_path == "z3":
            # solve the problem
            start = datetime.datetime.now()
            result = self.optimizer.check()
            solve_time = datetime.datetime.now() - start
            assert str(result) == "sat", "No solution found"

            # validate the model
            model = self.optimizer.model()
            assert self.validate_model(model, lights), "Model is invalid"
            assert self.switch_vars is not None
            switches = [1 if model[var] else 0 for var in self.switch_vars]
        else:
            self.optimizer.set("pp.wcnf", True)
            wcnf = str(self.optimizer)
            # Note: This merely calls the solver. It does not interpret the output.
            #       This is just to measure the time it takes to solve the problem.
            with Path("./solver-out_" + solver_path.split("/")[-1] + ".txt").open("a+") as out:
                start = datetime.datetime.now()
                subprocess.run([solver_path, wcnf], stdout=out, check=False)
                solve_time = datetime.datetime.now() - start

        # pop the context from the optimizer
        self.optimizer.pop()

        return switches, constr_time, solve_time


def simulate_error_rate(code: ColorCode, error_rate: float, nr_sims: int, solver_path: str = "z3") -> dict[str, Any]:
    """Simulate the logical error rate for a given distance and error rate."""
    problem = LightsOut(code.faces_to_qubits, code.qubits_to_faces)

    start = datetime.datetime.now()
    problem.preconstruct_z3_instance()
    preconstr_time = datetime.datetime.now() - start
    min_wt_logicals: npt.NDArray[np.int_] = np.full(len(code.L), -1).astype(int)
    logical_errors: npt.NDArray[np.int_] = np.zeros(len(code.L)).astype(int)
    avg_constr_time = 0.0
    avg_solve_time = 0.0
    rng = np.random.default_rng()
    for i in range(nr_sims):
        # sample error
        error = rng.choice([0, 1], size=code.n, p=[1 - error_rate, error_rate])

        # get syndrome
        syndrome = code.get_syndrome(error)
        lights = [bool(b) for b in syndrome]

        # compute estimate
        estimate, constr_time, solve_time = problem.solve(lights, solver_path=solver_path)
        if len(estimate) > 0:
            # check if the estimate is correct
            residual = (error + np.array(estimate)) % 2
            for logical in range(len(code.L)):
                if (code.L[logical] @ residual % 2).any():
                    logical_errors[logical] += 1
                    wt = np.sum(residual)  # compute the min weight of a logical error
                    if min_wt_logicals[logical] == -1 or wt < min_wt_logicals[logical]:
                        min_wt_logicals[logical] = int(wt)
                    break

        # compute rolling average of the times
        avg_constr_time = (avg_constr_time * i + constr_time.microseconds) / (i + 1)
        avg_solve_time = (avg_solve_time * i + solve_time.microseconds) / (i + 1)

    logical_error_rates: list[float] = [nr_errors / nr_sims for nr_errors in logical_errors]
    logical_error_rate_ebs: list[float] = [np.sqrt((1 - ler) * ler / nr_sims) for ler in logical_error_rates]
    avg_total_time = avg_constr_time + avg_solve_time

    return {
        "lattice": code.lattice_type,
        "distance": code.distance,
        "p": error_rate,
        "logical_error_rates": logical_error_rates,
        "logical_error_rate_ebs": logical_error_rate_ebs,
        "preconstr_time": preconstr_time.microseconds,
        "avg_constr_time": avg_constr_time,
        "avg_solve_time": avg_solve_time,
        "avg_total_time": avg_total_time,
        "min_wts_logical_err": min_wt_logicals.tolist(),
    }


def run(
    lattice_type: str,
    distance: int,
    error_rate: float,
    nr_sims: int = 10000,
    results_dir: str = "./results_maxsat",
    solver: str = "z3",
) -> None:
    """Run the decoding simulation for a given distance and error rate."""
    code = code_from_string(lattice_type, distance)
    data = simulate_error_rate(code, error_rate, nr_sims, solver)
    strg = solver.split("/")[-1]
    filename = f"./code={code.lattice_type},distance={code.distance},p={round(error_rate, 4)},solver={strg}.json"
    path = Path(results_dir)
    path.mkdir(parents=True, exist_ok=True)
    with (path / filename).open("w") as out:
        out.write(json.dumps(data))



"""A color code class."""

from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING

import numpy as np
from ldpc import mod2

if TYPE_CHECKING:  # pragma: no cover
    import numpy.typing as npt


class LatticeType(str, Enum):
    """The type of lattice used for the color code."""

    HEXAGON = "hexagon"
    SQUARE_OCTAGON = "square_octagon"


class ColorCode:
    """A base class for color codes on a three-valent, three-colourable lattice."""

    def __init__(self, distance: int, lattice_type: LatticeType) -> None:
        """Initialize the color code."""
        self.distance = distance
        self.ancilla_qubits: set[tuple[int, int]] = set()
        self.data_qubits: set[tuple[int, int]] = set()
        self.qubits_to_faces: dict[int, list[int]] = {}
        self.faces_to_qubits: dict[int, list[int]] = {}
        self.lattice_type = lattice_type
        self.add_qubits()
        self.H: npt.NDArray[np.int_] = np.zeros((len(self.ancilla_qubits), len(self.data_qubits)), dtype=int)
        self.construct_layout()
        self.compute_logical()
        self.n = len(self.qubits_to_faces)
        self.k = self.L.shape[1]

    def __eq__(self, other: object) -> bool:
        """Check if two color codes are equal."""
        if not isinstance(other, ColorCode):
            return NotImplemented
        return np.array_equal(self.H, other.H)

    def add_qubits(self) -> None:
        """Compute the ancilla and data qubits lists from the lattice type."""

    def construct_layout(self) -> None:
        """Construct the adjacency lists of the code from the qubits lists. Assumes add_qubits was called."""

    def compute_logical(self) -> None:
        """Compute the logical matrix L."""
        ker_hx = mod2.nullspace(self.H)  # compute the kernel basis of hx
        im_hz_transp = mod2.row_basis(self.H)  # compute the image basis of hz.T
        log_stack = np.vstack([im_hz_transp, ker_hx])
        pivots = mod2.row_echelon(log_stack.T)[3]
        log_op_indices = [i for i in range(im_hz_transp.shape[0], log_stack.shape[0]) if i in pivots]
        self.L = log_stack[log_op_indices]

    def get_syndrome(self, error: npt.NDArray[np.int_]) -> npt.NDArray[np.int_]:
        """Compute the syndrome of the error."""
        return self.H @ error % 2

    def check_if_logical_error(self, residual: npt.NDArray[np.int_]) -> bool:
        """Check if the residual is a logical error."""
        return (self.L @ residual % 2).any() is True



"""Square Octagon Color Code. Created by Peter-Jan Derks."""
from __future__ import annotations

from .color_code import ColorCode, LatticeType


class SquareOctagonColorCode(ColorCode):
    """4.8.8 triangular colour code."""

    def __init__(self, distance: int) -> None:
        """4.8.8 triangular colour code.

        This class can be used to generate the parity check matrix of 4.8.8 triangular colour code.
        This code has parameters [n, k, d] = [1/2*d**2 + d - 1/2, 1, d].

        param:  distance: Distance of the code to generate. Must be an odd integer.
        """
        # additionally to ancilla_qubits (on squares) we have the ones on octagons
        self.octagon_ancilla_qubits: set[tuple[int, int]] = set()
        self.square_ancilla_qubits: set[tuple[int, int]] = set()
        ColorCode.__init__(self, distance=distance, lattice_type=LatticeType.SQUARE_OCTAGON)

    def add_qubits(self) -> None:
        """Add qubits to the code."""
        self.bottom_row_ancillas()
        y = 1
        x_max = self.distance

        while y <= (self.distance + self.distance // 2):
            self.even_data_qubit_row(x_max, y)
            if y == 1:
                x_max -= 1
            else:
                x_max -= 2
            y += 1
            if y <= (self.distance + self.distance // 2):
                self.odd_data_qubit_row(x_max, y)
                y += 1
            if y <= (self.distance + self.distance // 2):
                if y % 2 == 0:
                    self.even_ancilla_qubit_row(x_max, y)
                else:
                    self.odd_ancilla_qubit_row(x_max, y)
                y += 1

        self.ancilla_qubits = self.square_ancilla_qubits.union(self.octagon_ancilla_qubits)

    def even_ancilla_qubit_row(self, x_max: int, y: int) -> None:
        """Create even ancilla qubits."""
        x = y + 1
        n_qubits = 0
        while True:
            if n_qubits >= x_max:
                break
            self.square_ancilla_qubits.add((x, y))
            n_qubits += 1
            x += 3
            if n_qubits >= x_max:
                break
            self.octagon_ancilla_qubits.add((x, y))
            n_qubits += 1
            x += 3

    def odd_ancilla_qubit_row(self, x_max: int, y: int) -> None:
        """Create odd ancilla qubits."""
        x = y - 2
        n_qubits = 0
        while True:
            if n_qubits >= x_max:
                break
            self.octagon_ancilla_qubits.add((x, y))
            n_qubits += 1
            x += 3
            if n_qubits >= x_max:
                break
            self.square_ancilla_qubits.add((x, y))
            n_qubits += 1
            x += 3

    def even_data_qubit_row(self, x_max: int, y: int) -> None:
        """Create even data qubits."""
        x = y - 1
        n_qubits = 0
        while True:
            if n_qubits >= x_max:
                break
            self.data_qubits.add((x, y))
            n_qubits += 1
            x += 2
            if n_qubits >= x_max:
                break
            self.data_qubits.add((x, y))
            n_qubits += 1
            x += 4

    def odd_data_qubit_row(self, x_max: int, y: int) -> None:
        """Create odd data qubits."""
        x = y + 1
        n_qubits = 0
        while True:
            if n_qubits >= x_max:
                break
            self.data_qubits.add((x, y))
            n_qubits += 1
            x += 2
            if n_qubits >= x_max:
                break
            self.data_qubits.add((x, y))
            n_qubits += 1
            x += 4

    def bottom_row_ancillas(self) -> None:
        """Create ancilla qubits on the bottom row of the lattice."""
        for x in range(4, self.distance // 2 * 6, 6):
            self.octagon_ancilla_qubits.add((x, 0))

    def construct_layout(self) -> None:
        """Construct the layout of the code."""
        coords_to_idx: dict[tuple[int, int], int] = {}
        # builds a map: {(x,y): index} for each qubit with coordinates (x,y)
        # initializes the {qubit_index: [faces]} adjacency list
        for idx, q in enumerate(self.data_qubits):
            coords_to_idx[q] = idx
            self.qubits_to_faces[idx] = []

        for idx, (x, y) in enumerate(self.square_ancilla_qubits):
            qbts: list[int] = []
            for coord in [(x - 1, y - 1), (x + 1, y - 1), (x + 1, y + 1), (x - 1, y + 1)]:
                if coord in coords_to_idx:
                    qubit_idx = coords_to_idx[coord]
                    qbts.append(qubit_idx)
                    self.qubits_to_faces[qubit_idx].append(idx)

            self.faces_to_qubits[idx] = qbts
            for qb in qbts:
                self.H[idx, qb] = 1

        for idx, (x, y) in enumerate(self.octagon_ancilla_qubits):
            idx += len(self.square_ancilla_qubits)  # noqa: PLW2901
            qbts = []
            for coord in [
                (x - 2, y - 1),
                (x - 1, y - 2),
                (x + 1, y - 2),
                (x + 2, y - 1),
                (x + 2, y + 1),
                (x + 1, y + 2),
                (x - 1, y + 2),
                (x - 2, y + 1),
            ]:
                if coord in coords_to_idx:
                    qubit_idx = coords_to_idx[coord]
                    qbts.append(qubit_idx)
                    self.qubits_to_faces[qubit_idx].append(idx)

            self.faces_to_qubits[idx] = qbts
            for qb in qbts:
                self.H[idx, qb] = 1



"""Plotting functions for the paper."""
from __future__ import annotations

import argparse
import json
import os
from pathlib import Path
from typing import TYPE_CHECKING, Any, cast

import numpy as np
from matplotlib import pyplot as plt

if TYPE_CHECKING:
    from matplotlib.axes import Axes
from scipy.optimize import curve_fit

ler_k = "logical_error_rates"
ler_eb_k = "logical_error_rate_ebs"
min_wts_k = "min_wts_logical_err"


def plot_ler_vs_distance(code_dict: dict[float, Any], ax: Axes, pers: list[float], logical_idx: int = 0) -> None:
    """Plot the logical error rate vs distance for different err rates."""
    for p in pers:
        ds = []
        lers = []
        for d, ler in sorted(code_dict[p].items()):
            ds.append(int(d))
            lers.append(ler[logical_idx])
        ax.plot(ds, lers, label="p=" + str(round(p, 2)), marker="o", linestyle="-")
    ax.set_yscale("log")
    ax.legend()
    ax.set_ylabel("Logical failure rate")
    ax.set_xlabel(r"Code distance $\it{d}$")


def threshold_fit(variables: tuple[float, float], b0: float, b1: float, b2: float, mu: float, pth: float) -> float:
    """Compute standard fit function for the threshold."""
    p, ell = variables
    expr = (p - pth) * (ell ** (1 / mu))
    return cast(float, b0 + b1 * expr + b2 * (expr**2))


def calculate_threshold(
    ax: Axes,
    code_dict: dict[int, Any],
    min_per: float = 0.06,
    max_per: float = 0.13,
    title: str | None = None,
    logical_idx: int = 0,
) -> None:
    """Calculate the threshold for the given results."""
    ler_data = []
    ler_eb_data = []
    per_data = []
    distance_data = []
    for distance in code_dict:
        per_array = []
        ler_array = []
        ler_eb = []

        for index, per in enumerate(code_dict[distance]["p"]):
            if min_per < per < max_per:
                per_array.append(per)
                ler_array.append(code_dict[distance][ler_k][index][logical_idx])
                ler_eb.append(code_dict[distance][ler_eb_k][index][logical_idx])
        per_data.extend(per_array)
        ler_data.extend(ler_array)
        ler_eb_data.extend(ler_eb)
        distance_data.extend([int(distance) for _ in range(len(per_array))])

    popt, pcov = curve_fit(threshold_fit, (per_data, distance_data), ler_data, maxfev=10000)
    if ax is not None:
        ax.axvline(x=popt[-1], color="black", linestyle="dashed")
        print("threshold: ", popt[-1])

    distance_array = [int(distance) for distance in code_dict]
    distance_array.sort()
    for distance in distance_array:
        per_array = code_dict[distance]["p"]
        la = np.array(code_dict[distance][ler_k])
        ler_arr = la[:, logical_idx]
        ea = np.array(code_dict[distance][ler_eb_k])
        ler_eb_ar = ea[:, logical_idx]

        if per_array != [] and ax is not None:
            ax.errorbar(per_array, ler_arr, yerr=ler_eb_ar, label="d = " + str(distance), fmt="|")

    ax.legend()
    ax.set_xlabel("Physical error rate")
    ax.set_ylabel("Logical error rate")
    ax.set_title(title)
    ax.set_yscale("log")
    ax.set_xlim(min_per, max_per)


def generate_plots(results_dir: Path, results_file: Path) -> None:
    """Generate the plots for the paper."""
    # read in all generated data
    data = []
    for file in results_dir.glob("*.json"):
        with file.open() as f:
            data.append(json.loads(f.read()))

    fig, ax = plt.subplots(4, 4, figsize=(12, 12))
    metrics: dict[int, Any] = {}
    per_metrics: dict[float, Any] = {}

    for result in data:
        d = result["distance"]
        p = result["p"]

        if d not in metrics:
            metrics[d] = {
                "p": [],
                ler_k: [],
                ler_eb_k: [],
                "avg_total_time": [],
                min_wts_k: [],
            }
        if p not in per_metrics:
            per_metrics[p] = {}

        metrics[d]["p"].append(p)
        metrics[d][ler_k].append(result[ler_k])
        metrics[d][ler_eb_k].append(result[ler_eb_k])
        metrics[d]["avg_total_time"].append(result["avg_total_time"])
        metrics[d][min_wts_k].append(result[min_wts_k])

        per_metrics[p][d] = result[ler_k]
    for d, mdata in sorted(metrics.items()):
        (mdata["p"], mdata[ler_k], mdata["avg_total_time"], mdata[ler_eb_k]) = zip(
            *sorted(zip(mdata["p"], mdata[ler_k], mdata["avg_total_time"], mdata[ler_eb_k]))
        )

        # sum over all logical to get overall ler+ebs
        mdata["ler_sum"] = np.sum(mdata[ler_k], axis=1)
        mdata["ler_eb_sum"] = np.sum(mdata[ler_eb_k], axis=1)

        # plot lers for different logicals separately
        for logical_idx in range(len(mdata[ler_k][0])):
            y = np.array(mdata[ler_k])
            y = y[:, logical_idx]
            ebs = np.array(mdata[ler_eb_k])
            ebs = ebs[:, logical_idx]
            ax[0][logical_idx].errorbar(mdata["p"], y, ebs, label="d=" + str(d))
        ax[0][0].set_xlabel("Physical error rate")
        ax[0][0].set_ylabel("Logical error rate")
        ax[0][0].legend()

        ax[1][0].plot(mdata["p"], mdata["avg_total_time"], label="d=" + str(d))
        ax[1][0].set_xlabel("Physical error rate")
        ax[1][0].set_ylabel("Average time per run (µs)")
        ax[1][0].legend()
        ax[1][0].set_ylim(0, 185000)

    # plot the average time per run over the distance
    ds = []
    p_data: dict[float, Any] = {}
    pers = [0.001, 0.02, 0.05, 0.08, 0.13]
    for d, mdata in sorted(metrics.items()):
        ds.append(d)
        for i, p in enumerate(mdata["p"]):
            if p in pers:
                if p not in p_data:
                    p_data[p] = {"d": [], "t": []}
                p_data[p]["d"].append(d)
                p_data[p]["t"].append(mdata["avg_total_time"][i])
    for p, pdata in sorted(p_data.items()):
        ax[2][1].plot(ds, pdata["t"], label="p=" + str(p))

    ax[1][1].set_xlabel("Distance")
    ax[1][1].set_ylabel("Average time per run (µs)")
    ax[1][1].legend()
    ax[1][1].set_xticks(ds)
    ax[1][1].set_ylim(0, 185000)
    for logical_idx in range(len(mdata[ler_k][0])):
        calculate_threshold(code_dict=metrics, ax=ax[1][logical_idx], title="Threshold")
    plot_ler_vs_distance(per_metrics, ax=ax[3][0], pers=[0.01])
    # save plot as vector graphic
    plt.savefig(results_file, bbox_inches="tight")


def generate_plots_tn(results_dir: Path, results_file: Path) -> None:
    """Generate the plots for the tensor network decoder."""
    # read in all generated data
    data = []
    for file in results_dir.glob("*.json"):
        with file.open() as f:
            data.append(json.loads(f.read()))

    # prepare code to per,ler map and print
    code_to_xys: dict[float, Any] = {}
    for run in data:
        xys = code_to_xys.setdefault(run["n_k_d"][-1], [])
        xys.append((run["physical_error_rate"], run["logical_failure_rate"]))

    for xys in code_to_xys.values():
        xys.sort(key=lambda xy: xy[0])

    fig, ax = plt.subplots(2, 2, figsize=(12, 10))
    # add data
    for code, xys in sorted(code_to_xys.items()):
        ax[0][0].plot(*zip(*xys), "x-", label=f"d={code}")
    ax[0][0].set_xlabel("Physical error rate")
    ax[0][0].set_ylabel("Logical error rate")
    ax[0][0].legend()

    # prepare code to per,average-time-per-run map and print
    code_to_xys = {}
    for run in data:
        xys = code_to_xys.setdefault(run["n_k_d"][-1], [])
        # convert from seconds to microseconds
        xys.append((run["error_probability"], (run["wall_time"] / run["n_run"]) * 1e6))

    for xys in code_to_xys.values():
        xys.sort(key=lambda xy: xy[0])

    for code, xys in sorted(code_to_xys.items()):
        ax[1][0].plot(*zip(*xys), "x-", label=f"d={code}")
    ax[1][0].set_xlabel("Physical error rate")
    ax[1][0].set_ylabel("Average time per run (µs)")
    ax[1][0].legend()
    ax[1][0].set_ylim(0, 300000)

    ds = []
    p_data: dict[float, Any] = {}
    pers = [0.001, 0.021, 0.051, 0.081, 0.111]
    for d, cdata in sorted(code_to_xys.items()):
        ds.append(d)
        for _, (p, t) in enumerate(cdata):
            if p in pers:
                if p not in p_data:
                    p_data[p] = {"d": [], "t": []}
                p_data[p]["d"].append(d)
                p_data[p]["t"].append(t)
    for p, pdata in sorted(p_data.items()):
        ax[1][1].plot(ds, pdata["t"], label="p=" + str(p))

    ax[1][1].set_xlabel("Distance")
    ax[1][1].set_ylabel("Average time per run (µs)")
    ax[1][1].legend()
    # ax[1][1].set_yscale("log")
    ax[1][1].set_xticks(ds)
    ax[1][1].set_ylim(0, 300000)
    # save plot as vector graphic
    plt.savefig(results_file, bbox_inches="tight")


def generate_plots_comp(results_dir: Path, results_file: Path) -> None:
    """Generate plots for the comparison of the different solvers."""
    fig, ax = plt.subplots(2, figsize=(12, 12))
    cols = ["blue", "orange", "green", "red", "purple", "brown", "pink", "gray", "cyan", "olive"]
    idx = 0
    solver_to_col: dict[str, str] = {}
    p_to_col: dict[float, str] = {}
    for subdir, _, files in os.walk(results_dir):
        if not files:
            continue
        data = []
        solver = subdir.split("/")[-1]
        if solver not in solver_to_col:
            solver_to_col[solver] = cols[idx]
            idx += 1
        for f in files:
            fp = subdir + "/" + f
            with Path(fp).open() as ff:
                data.append(json.loads(ff.read()))

        metrics: dict[int, dict[str, list[Any]]] = {}
        for result in data:
            d = result["distance"]
            p = result["p"]
            if d not in metrics:
                metrics[d] = {"p": [], "avg_total_time": []}

            metrics[d]["p"].append(p)
            metrics[d]["avg_total_time"].append(result["avg_total_time"])

        for d, mdata in sorted(metrics.items()):
            if d in (17, 21):
                (
                    mdata["p"],
                    mdata["avg_total_time"],
                ) = zip(*sorted(zip(mdata["p"], mdata["avg_total_time"])))
                ax[0].plot(
                    mdata["p"],
                    mdata["avg_total_time"],
                    label="solver=" + solver + ", d=" + str(d),
                    color=solver_to_col[solver],
                )
        ds: list[int] = []
        idx = 0
        p_data: dict[float, dict[str, list[Any]]] = {}
        pers = [0.001, 0.051, 0.131]
        for d, pdata in sorted(metrics.items()):
            ds.append(d)
            for i, p in enumerate(pdata["p"]):
                if p in pers:
                    if p not in p_data:
                        p_data[p] = {"d": [], "t": []}
                        p_to_col[p] = cols[idx]
                        idx += 1
                    p_data[p]["d"].append(d)
                    p_data[p]["t"].append(pdata["avg_total_time"][i])
        for p, ppdata in sorted(p_data.items()):
            ax[1].plot(
                ds,
                ppdata["t"],
                label="p=" + str(p) + ", " + (solver if solver == "z3" else "CASHWMaxSAT-CorePlus"),
                color=p_to_col[p],
                marker="x" if solver == "z3" else "o",
            )

        ax[1].set_xlabel("Distance")
        ax[1].set_ylabel("Average time per run (µs)")
        ax[1].legend()
        ax[1].set_xticks(ds)

        ax[0].set_xlabel("Physical error rate")
        ax[0].set_ylabel("Average time per run (µs)")
        ax[0].legend(loc="center left", bbox_to_anchor=(1, 0.5))

    # save plot as vector graphic
    plt.savefig(results_file, bbox_inches="tight")


def main() -> None:
    """Run main CLI function."""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "results_dir",
        type=str,
        help="Path to the directory containing the results",
    )
    parser.add_argument(
        "--mode",
        type=str,
        default="maxsat",
        help="Mode to use. Either 'maxsat' or 'tn'. Default: maxsat",
    )
    parser.add_argument(
        "--results_file",
        type=str,
        default="./results.pdf",
        help="File where the results should be saved. Default: ./results.pdf",
    )
    args = parser.parse_args()

    if args.mode == "maxsat":
        generate_plots(Path(args.results_dir), Path(args.results_file))
    elif args.mode == "tn":
        generate_plots_tn(Path(args.results_dir), Path(args.results_file))
    elif args.mode == "comparison":
        generate_plots_comp(Path(args.results_dir), Path(args.results_file))
    else:
        msg = "Unknown mode: " + args.mode
        raise ValueError(msg)



"""Command line interface for the CC decoder."""
from __future__ import annotations

import argparse

from . import decoder, tn_decoder


def cli() -> None:
    """Run the CC decoder as cli."""
    parser = argparse.ArgumentParser()
    parser.add_argument("distance", type=int, help="the distance of the code")
    parser.add_argument("error_rate", type=float, help="the error rate")
    parser.add_argument(
        "--type",
        type=str,
        default="hexagon",
        help="type of the code lattice (hexagon or square_octagon). Default: hexagon",
    )
    parser.add_argument("--nr_sims", type=int, default=10000, help="the number of simulations to run. Default: 10000")
    parser.add_argument(
        "--results_dir", type=str, default="./results", help="the directory to save the results to. Default: ./results"
    )
    parser.add_argument(
        "--decoder", type=str, default="maxsat", help="the decoder to use (maxsat or tn). Default: maxsat"
    )

    parser.add_argument(
        "--solver", type=str, default="z3", help="maxsat solver to use (path to a executable). Default: z3"
    )

    args = parser.parse_args()

    if args.decoder == "maxsat":
        decoder.run(args.type, args.distance, args.error_rate, args.nr_sims, args.results_dir, args.solver)
    elif args.decoder == "tn":
        tn_decoder.run(args.distance, args.error_rate, args.nr_sims, args.results_dir)
    else:
        msg = f"Unknown decoder {args.decoder}. Please choose either maxsat or tn."
        raise ValueError(msg)



"""MQT Core - The Backbone of the Munich Quantum Toolkit."""

from __future__ import annotations

from ._core import add
from ._version import version as __version__

__all__ = ["__version__", "add"]



"""Contains the implementation of the tensor network decoder for the hexagonal color code."""
from __future__ import annotations

from .color_code import ColorCode, LatticeType
from .comparison import tn_decoder
from .hexagonal_color_code import HexagonalColorCode
from .square_octagon_color_code import SquareOctagonColorCode


def code_from_string(lattice_type: str, distance: int) -> ColorCode:
    """Construct a color code from a string defining the lattice and a distance."""
    if lattice_type == LatticeType.HEXAGON:
        return HexagonalColorCode(distance)
    if lattice_type == LatticeType.SQUARE_OCTAGON:
        return SquareOctagonColorCode(distance)
    msg = f"Unknown lattice type {lattice_type}. Please choose either hexagon or square_octagon."
    raise ValueError(msg)


__all__ = [
    "tn_decoder",
    "HexagonalColorCode",
    "ColorCode",
    "LatticeType",
    "SquareOctagonColorCode",
    "code_from_string",
]



"""Construction of HGP codes."""
from __future__ import annotations

import numpy as np
from bposd.hgp import hgp

# med sized HGP code from

h = np.loadtxt("./mkmn_24_6_10.txt").astype(int)
seed_code = np.loadtxt("./mkmn_24_6_10.txt").astype(int)
# print(seed_code)
qcode = hgp(seed_code, compute_distance=True)
qcode.canonical_logicals()
qcode.test()
print(qcode.code_params)
print(qcode.hx)
print("hx:")
print(qcode.hz)
np.savetxt(f"./hgp_{qcode.code_params}_hz.txt", qcode.hz, fmt="%d", newline="\n")
np.savetxt(f"./hgp_{qcode.code_params}_hx.txt", qcode.hx, fmt="%d", newline="\n")

# larger code
# a1 = pt.array([
#     [(0), (11), (7), (12)],
#     [(1), (8), (1), (8)],
#     [(11), (0), (4), (8)],
#     [(6), (2), (4), (12)]])
# H = a1.to_binary(lift_parameter=13)
# qcode = hgp(H, H, compute_distance=True)
# qcode.test()
# np.savetxt(f"./hgp_{qcode.code_params}_hz.txt", qcode.hz, fmt='%d', newline='\n')



"""Decoding simulation using the tensor network implementation of the qecsim package."""
from __future__ import annotations

import json
from pathlib import Path

from qecsim import app
from qecsim.models.color import Color666Code, Color666MPSDecoder
from qecsim.models.generic import BitFlipErrorModel


def run(distance: int, error_rate: float, nr_sims: int = 10000, results_dir: str = "./results_tn") -> None:
    """Run the decoder for the hexagonal color code.

    :param distance: distance to run
    :param error_rate: error rate to run
    :param nr_sims: number of samples to run
    :param results_dir: directory to store results.
    """
    code = Color666Code(distance)
    error_model = BitFlipErrorModel()
    decoder = Color666MPSDecoder(chi=8)
    data = app.run(code, error_model, decoder, error_rate, max_runs=nr_sims)
    filename = f"distance={distance},p={round(error_rate, 4)}.json"
    path = Path(results_dir)
    path.mkdir(parents=True, exist_ok=True)
    with (path / filename).open("w") as out:
        out.write(json.dumps(data))



"""Hexagonal Color Code class.

Created by Peter-Jan Derks
Hexagonal Color Code layout construction adapted from https://github.com/peter-janderks/restriction_decoder_domain_wall_colour_code
"""

from __future__ import annotations

from .color_code import ColorCode, LatticeType


class HexagonalColorCode(ColorCode):
    """Hexagonal Color Code."""

    def __init__(self, distance: int) -> None:
        """Hexagonal Color Code initialization from base class."""
        ColorCode.__init__(self, distance=distance, lattice_type=LatticeType.HEXAGON)

    def add_qubits(self) -> None:
        """Add qubits to the code."""
        colour = ["r", "b", "g"]
        y = 0
        x_max = self.distance + self.distance // 2
        while x_max > 0:
            ancilla_colour = colour[y % 3]
            if ancilla_colour == "r":
                self.red_row(x_max, y)
            elif ancilla_colour == "b":
                self.blue_row(x_max, y)
            else:
                self.green_row(x_max, y)
            x_max -= 1
            y += 1

    def red_row(self, x_max: int, y: int) -> None:
        """Create red ancilla qubits."""
        i = 0
        x_row = y
        while i < x_max:
            data_or_ancilla = i % 3
            if data_or_ancilla in (0, 2):
                self.data_qubits.add((x_row, y))
            else:
                self.ancilla_qubits.add((x_row, y))
            i += 1
            x_row += 2

    def blue_row(self, x_max: int, y: int) -> None:
        """Create blue ancilla qubits."""
        i = 0
        x_row = y
        while i < x_max:
            data_or_ancilla = i % 3
            if data_or_ancilla in (0, 1):
                self.data_qubits.add((x_row, y))
            else:
                self.ancilla_qubits.add((x_row, y))
            i += 1
            x_row += 2

    def green_row(self, x_max: int, y: int) -> None:
        """Create green ancilla qubits."""
        i = 0
        x_row = y
        while i < x_max:
            data_or_ancilla = i % 3
            if data_or_ancilla in (1, 2):
                self.data_qubits.add((x_row, y))
            else:
                self.ancilla_qubits.add((x_row, y))
            x_row += 2
            i += 1

    def construct_layout(self) -> None:
        """Construct the layout of the hexagonal color code."""
        coords_to_idx: dict[tuple[int, int], int] = {}
        # builds a map: {(x,y): index} for each qubit with coordinates (x,y)
        # initializes the {qubit_index: [faces]} adjacency list
        for idx, q in enumerate(self.data_qubits):
            coords_to_idx[q] = idx
            self.qubits_to_faces[idx] = []

        for idx, (x, y) in enumerate(self.ancilla_qubits):
            qbts: list[int] = []
            for coord in [(x - 1, y + 1), (x - 2, y), (x - 1, y - 1), (x + 1, y - 1), (x + 2, y), (x + 1, y + 1)]:
                if coord in coords_to_idx:
                    qubit_idx = coords_to_idx[coord]
                    qbts.append(qubit_idx)
                    self.qubits_to_faces[qubit_idx].append(idx)
            self.faces_to_qubits[idx] = qbts
            for qb in qbts:
                self.H[idx, qb] = 1

        # L is the matrix of logicals of the code
        self.compute_logical()



"""Generate the lifted hypergraph product of the protographs a and b. From Roffe's LDPC library."""  # noqa: EXE002
from __future__ import annotations

from typing import TYPE_CHECKING

# taken from https://github.com/quantumgizmos/bias_tailored_qldpc
import ldpc.protograph as pt
import numpy as np

if TYPE_CHECKING:
    import numpy.typing as npt
from bposd.css import css_code
from bposd.stab import stab_code


def identity(n: int) -> npt.NDArray[np.int_]:
    """Return the identity matrix of size n."""
    return pt.identity(n)


class LiftedHgp(css_code):
    """Lifted Hypergraph Product code."""

    def __init__(self, lift_parameter: int, a: npt.NDArray[np.int_], b: npt.NDArray[np.int_] | None = None) -> None:
        """Generate the lifted hypergraph product of the protographs a and b."""
        self.a = a

        self.a_m, self.a_n = self.a.shape

        if b is None:
            self.b = pt.copy(self.a)
        else:
            self.b = b

        self.b_m, self.b_n = self.b.shape

        self.hx1_proto = np.kron(self.a, identity(self.b_n))
        self.hx2_proto = np.kron(identity(self.a_m), self.b.T)
        self.hx_proto = pt.hstack([self.hx1_proto, self.hx2_proto])

        self.hz1_proto = np.kron(identity(self.a_n), self.b)
        self.hz2_proto = np.kron(self.a.T, identity(self.b_m))
        self.hz_proto = pt.hstack([self.hz1_proto, self.hz2_proto])

        self.lift_parameter = lift_parameter

        super().__init__(self.hx_proto.to_binary(lift_parameter), self.hz_proto.to_binary(lift_parameter))

    @property
    def protograph(self) -> npt.NDArray[np.int_]:
        """Returns the protograph of the lifted hypergraph product."""
        px = pt.vstack([pt.zeros(self.hz_proto.shape), self.hx_proto])
        pz = pt.vstack([self.hz_proto, pt.zeros(self.hx_proto.shape)])
        return pt.hstack([px, pz])

    @property
    def hx1(self) -> npt.NDArray[np.int_]:
        """Returns the first horizontal protograph of the lifted hypergraph product."""
        return self.hx1_proto.to_binary(self.lift_parameter)

    @property
    def hx2(self) -> npt.NDArray[np.int_]:
        """Returns the second horizontal protograph of the lifted hypergraph product."""
        return self.hx2_proto.to_binary(self.lift_parameter)

    @property
    def hz1(self) -> npt.NDArray[np.int_]:
        """Returns the first vertical protograph of the lifted hypergraph product."""
        return self.hz1_proto.to_binary(self.lift_parameter)

    @property
    def hz2(self) -> npt.NDArray[np.int_]:
        """Returns the second vertical protograph of the lifted hypergraph product."""
        return self.hz2_proto.to_binary(self.lift_parameter)


class BiasTailoredLiftedProduct(stab_code):
    """Generate the bias-tailored lifted hypergraph product of the protographs a and b. From Roffe's LDPC library."""

    def __init__(self, lift_parameter: int, a: npt.NDArray[np.int_], b: npt.NDArray[np.int_] | None = None) -> None:
        """Generate the bias-tailored lifted hypergraph product of the protographs a and b."""
        lhgp = LiftedHgp(lift_parameter, a, b)

        # Hadamard rotation
        temp1 = pt.hstack([pt.zeros(lhgp.hx1_proto.shape), lhgp.hz2_proto])
        temp2 = pt.hstack([lhgp.hx1_proto, pt.zeros(lhgp.hz2_proto.shape)])
        self.hx_proto = pt.vstack([temp1, temp2])
        temp1 = pt.hstack([lhgp.hz1_proto, pt.zeros(lhgp.hx2_proto.shape)])
        temp2 = pt.hstack([pt.zeros(lhgp.hz1_proto.shape), lhgp.hx2_proto])
        self.hz_proto = pt.vstack([temp1, temp2])

        super().__init__(self.hx_proto.to_binary(lift_parameter), self.hz_proto.to_binary(lift_parameter))

    @property
    def protograph(self) -> npt.NDArray[np.int_]:
        """Returns the protograph of the lifted hypergraph product."""
        px = pt.vstack([pt.zeros(self.hz_proto.shape), self.hx_proto])
        pz = pt.vstack([self.hz_proto, pt.zeros(self.hx_proto.shape)])
        return pt.hstack([px, pz])



from typing import Any, ClassVar, overload

class Code:
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, arg0: str) -> None: ...
    @overload
    def __init__(self, arg0: str, arg1: str) -> None: ...
    @overload
    def __init__(self, arg0: list[list[bool]]) -> None: ...
    @overload
    def __init__(self, arg0: list[list[bool]], arg1: list[list[bool]]) -> None: ...
    def get_hx(self) -> list[list[bool]]: ...
    def get_hz(self) -> list[list[bool]]: ...
    def get_syndrome(self, arg0: list[bool], arg1: list[bool]) -> list[bool]: ...
    def get_x_syndrome(self, arg0: list[bool]) -> list[bool]: ...
    @overload
    def is_stabilizer(self, arg0: list[bool]) -> bool: ...
    @overload
    def is_stabilizer(self, arg0: list[bool], arg1: list[bool]) -> bool: ...
    def is_x_stabilizer(self, arg0: list[bool]) -> bool: ...
    def json(self) -> dict[str, Any]: ...
    def set_hx(self, arg0: list[list[bool]]) -> None: ...
    def set_hz(self, arg0: list[list[bool]]) -> None: ...

    d: int
    k: int
    n: int

class Decoder:
    def __init__(self) -> None: ...
    def decode(self, arg0: list[bool]) -> None: ...
    def set_code(self, arg0: Code) -> None: ...
    def set_growth(self, arg0: GrowthVariant) -> None: ...

    growth: GrowthVariant
    result: DecodingResult

class DecoderType:
    __members__: ClassVar[dict[DecoderType, int]] = ...  # read-only
    original_uf: ClassVar[DecoderType] = ...
    uf_heuristic: ClassVar[DecoderType] = ...

    @overload
    def __init__(self, value: int) -> None: ...
    @overload
    def __init__(self, arg0: str) -> None: ...
    @overload
    def __init__(self, arg0: DecoderType) -> None: ...
    def __eq__(self, other: object) -> bool: ...
    def __getstate__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: object) -> bool: ...
    def __setstate__(self, state: int) -> None: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

class DecodingResult:
    def __init__(self) -> None: ...
    def json(self) -> dict[str, Any]: ...

    decoding_time: int
    estim_vec_idxs: list[int]
    estimate: list[bool]

class DecodingResultStatus:
    __members__: ClassVar[dict[DecodingResultStatus, int]] = ...  # read-only
    success: ClassVar[DecodingResultStatus] = ...
    failure: ClassVar[DecodingResultStatus] = ...

    @overload
    def __init__(self, value: int) -> None: ...
    @overload
    def __init__(self, arg0: str) -> None: ...
    @overload
    def __init__(self, arg0: DecodingResultStatus) -> None: ...
    def __eq__(self, other: object) -> bool: ...
    def __getstate__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: object) -> bool: ...
    def __setstate__(self, state: int) -> None: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

class DecodingRunInformation:
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, arg0: float, arg1: int, arg2: list[bool], arg3: list[bool], arg4: DecodingResult) -> None: ...
    @overload
    def __init__(
        self,
        arg0: float,
        arg1: int,
        arg2: list[bool],
        arg3: list[bool],
        arg4: DecodingResultStatus,
        arg5: DecodingResult,
    ) -> None: ...
    def json(self) -> dict[str, Any]: ...

    code_size: int
    error: list[bool]
    physical_err_r: float
    result: DecodingResult
    status: DecodingResultStatus
    syndrome: list[bool]

class DecodingSimulator:
    def __init__(self) -> None: ...
    def simulate_avg_runtime(
        self, arg0: str, arg1: float, arg2: int, arg3: str, arg4: int, arg5: DecoderType
    ) -> None: ...
    def simulate_wer(
        self, arg0: str, arg1: float, arg2: float, arg3: int, arg4: Code, arg5: float, arg6: DecoderType
    ) -> None: ...

class GrowthVariant:
    __members__: ClassVar[dict[GrowthVariant, int]] = ...  # read-only
    all_components: ClassVar[GrowthVariant] = ...
    invalid_components: ClassVar[GrowthVariant] = ...
    single_smallest: ClassVar[GrowthVariant] = ...
    single_random: ClassVar[GrowthVariant] = ...
    single_qubit_random: ClassVar[GrowthVariant] = ...

    @overload
    def __init__(self, value: int) -> None: ...
    @overload
    def __init__(self, arg0: str) -> None: ...
    @overload
    def __init__(self, arg0: GrowthVariant) -> None: ...
    def __eq__(self, other: object) -> bool: ...
    def __getstate__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: object) -> bool: ...
    def __setstate__(self, state: int) -> None: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

class UFDecoder(Decoder):
    def __init__(self) -> None: ...
    def decode(self, arg0: list[bool]) -> None: ...
    @property
    def growth(self) -> GrowthVariant: ...
    @growth.setter
    def growth(self, arg0: GrowthVariant) -> None: ...
    @property
    def result(self) -> DecodingResult: ...
    @result.setter
    def result(self, arg0: DecodingResult) -> None: ...

class UFHeuristic(Decoder):
    def __init__(self) -> None: ...
    def decode(self, arg0: list[bool]) -> None: ...
    def reset(self) -> None: ...
    @property
    def growth(self) -> GrowthVariant: ...
    @growth.setter
    def growth(self, arg0: GrowthVariant) -> None: ...
    @property
    def result(self) -> DecodingResult: ...
    @result.setter
    def result(self, arg0: DecodingResult) -> None: ...

def apply_ecc(circuit_name: object, ecc_name: str, ecc_frequency: int = 100) -> dict[str, str]: ...
def sample_iid_pauli_err(arg0: int, arg1: float) -> list[bool]: ...



"""Simulate the decoding performance of the UF heuristic decoder."""
from __future__ import annotations

from pathlib import Path

from mqt import qecc

code = qecc.Code("../../examples/lp_(4,8)-[[1024,18,nan]]_hx.txt", "../../examples/lp_(4,8)-[[1024,18,nan]]_hz.txt")
code.K = 18
outpath = "./dp-sims-bindings.out"

runs_per_p = 1
curr_p = 0.00001
max_per = 0.00003
step_size = 0.00001
nr_failed_runs = 0
code_k = code.K

with Path(outpath).open("w") as outfile:
    while curr_p < max_per:
        nr_failed_runs = 0
        for _ in range(runs_per_p):
            decoder = qecc.UFHeuristic()
            decoder.set_code(code)
            err = qecc.sample_iid_pauli_err(code.N, curr_p)
            decoder.decode(code.get_x_syndrome(err))
            result = decoder.result.estimate
            if not code.is_stabilizer(result):
                nr_failed_runs += 1
        frac_failes = nr_failed_runs / runs_per_p
        wer = frac_failes / code_k
        res = str(curr_p) + ":" + str(wer) + "\n"
        print(res)
        outfile.write(res)
        curr_p += step_size
